/*=========================================================================
 *
 *  Copyright NumFOCUS
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *         http://www.apache.org/licenses/LICENSE-2.0.txt
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *=========================================================================*/
#ifndef itkNormalizedCorrelationTwoImageToOneImageMetric_hxx
#define itkNormalizedCorrelationTwoImageToOneImageMetric_hxx

#include "itkImageRegionConstIteratorWithIndex.h"
#include "itkNormalizedCorrelationTwoImageToOneImageMetric.h"

#

namespace itk {

template <typename TFixedImage, typename TMovingImage>
NormalizedCorrelationTwoImageToOneImageMetric<TFixedImage,
    TMovingImage>::NormalizedCorrelationTwoImageToOneImageMetric()
{
    m_SubtractMean = false;
}

template <typename TFixedImage, typename TMovingImage>
typename NormalizedCorrelationTwoImageToOneImageMetric<TFixedImage, TMovingImage>::MeasureType
NormalizedCorrelationTwoImageToOneImageMetric<TFixedImage, TMovingImage>::CalculateMeasure(int imageId) const
{

    using FixedIteratorType = itk::ImageRegionConstIteratorWithIndex<FixedImageType>;
    using MovingIteratorType = itk::ImageRegionConstIteratorWithIndex<MovingImageType>;
    using AccumulateType = typename NumericTraits<MeasureType>::AccumulateType;

    FixedImageConstPointer fixedImage;
    ChangeInformationFilterPointer filter;
    FixedImageMaskPointer fixedImageMask;
    FixedImageRegionType fixedImageRegion;
    MeasureType measure;

    typename FixedImageType::IndexType indexFixed;
    typename MovingImageType::IndexType indexMoving;
    typename Superclass::InputPointType inputPoint;

    if (imageId == 1) {
        this->m_Interpolator1->SetTransform(this->m_Transform1);
        this->m_Interpolator1->Initialize();

        fixedImage = this->m_FixedImage1;
        fixedImageMask = this->m_FixedImageMask1;
        fixedImageRegion = this->GetFixedImageRegion1();
        filter = this->m_Filter1;
    } else if (imageId == 2) {
        this->m_Interpolator2->SetTransform(this->m_Transform2);
        this->m_Interpolator2->Initialize();

        fixedImage = this->m_FixedImage2;
        fixedImageMask = this->m_FixedImageMask2;
        fixedImageRegion = this->GetFixedImageRegion2();
        filter = this->m_Filter2;
    } else {
        itkExceptionMacro(<< "Non supported image id.");
    }

    FixedIteratorType ti(fixedImage, fixedImageRegion);

    this->m_NumberOfPixelsCounted = 0;
    int numberOfPixelsVisited = 0;

    filter->Update();
    MovingImageConstPointer imageDRTIn = filter->GetOutput();
    //this->WriteImage(imageDRTIn, "", "");

    AccumulateType sff = NumericTraits<AccumulateType>::ZeroValue();
    AccumulateType smm = NumericTraits<AccumulateType>::ZeroValue();
    AccumulateType sfm = NumericTraits<AccumulateType>::ZeroValue();
    AccumulateType sf = NumericTraits<AccumulateType>::ZeroValue();
    AccumulateType sm = NumericTraits<AccumulateType>::ZeroValue();

    RealType movingValueMin = NumericTraits<RealType>::max();
    RealType fixedValueMin = NumericTraits<RealType>::max();
    RealType movingValueMax = NumericTraits<RealType>::NonpositiveMin();
    RealType fixedValueMax = NumericTraits<RealType>::NonpositiveMin();

    RealType maxX = NumericTraits<RealType>::NonpositiveMin();
    RealType maxY = NumericTraits<RealType>::NonpositiveMin();
    RealType maxZ = NumericTraits<RealType>::NonpositiveMin();
    RealType minX = NumericTraits<RealType>::max();
    RealType minY = NumericTraits<RealType>::max();
    RealType minZ = NumericTraits<RealType>::max();

    AccumulateType m_meanCurr_f = NumericTraits<AccumulateType>::ZeroValue();
    AccumulateType m_meanCurr_m = NumericTraits<AccumulateType>::ZeroValue();
    AccumulateType m_meanPrev_f = NumericTraits<AccumulateType>::ZeroValue();
    AccumulateType m_meanPrev_m = NumericTraits<AccumulateType>::ZeroValue();
    AccumulateType m_sPrev_f = NumericTraits<AccumulateType>::ZeroValue();
    AccumulateType m_sPrev_m = NumericTraits<AccumulateType>::ZeroValue();
    AccumulateType m_sCurr_f = NumericTraits<AccumulateType>::ZeroValue();
    AccumulateType m_sCurr_m = NumericTraits<AccumulateType>::ZeroValue();
    AccumulateType m_varianceCurr_f = NumericTraits<AccumulateType>::ZeroValue();
    AccumulateType m_varianceCurr_m = NumericTraits<AccumulateType>::ZeroValue();

    while (!ti.IsAtEnd()) {

        indexFixed = ti.GetIndex();
        ++numberOfPixelsVisited;

        fixedImage->TransformIndexToPhysicalPoint(indexFixed, inputPoint);

        if (fixedImageMask && !fixedImageMask->IsInsideInWorldSpace(inputPoint)) {
            ++ti;
            continue;
        }

        if (this->m_MovingImageMask && !this->m_MovingImageMask->IsInsideInWorldSpace(inputPoint)) {
            ++ti;
            continue;
        }

        const RealType fixedValue = ti.Get();

        if (fixedValue == -1024) {
            ++ti;
            continue;
        }

        if (imageDRTIn->TransformPhysicalPointToIndex(inputPoint, indexMoving)) {

            double x = inputPoint[0];
            double y = inputPoint[1];
            double z = inputPoint[2];
            maxX = maxX > x ? maxX : x;
            minX = minX < x ? minX : x;
            maxY = maxY > y ? maxY : y;
            minY = minY < y ? minY : y;
            maxZ = maxZ > z ? maxZ : z;
            minZ = minZ < z ? minZ : z;

            const RealType movingValue = imageDRTIn->GetPixel(indexMoving);

            if (NumericTraits<short>::NonpositiveMin() == movingValue) {
                ++ti;
                continue;
            }

            movingValueMin = movingValue < movingValueMin ? movingValue : movingValueMin;
            movingValueMax = movingValue > movingValueMax ? movingValue : movingValueMax;
            fixedValueMin = fixedValue < fixedValueMin ? fixedValue : fixedValueMin;
            fixedValueMax = fixedValue > fixedValueMax ? fixedValue : fixedValueMax;

            sff += fixedValue * fixedValue;
            smm += movingValue * movingValue;
            sfm += fixedValue * movingValue;
            if (this->m_SubtractMean) {
                sf += fixedValue;
                sm += movingValue;
            }
            this->m_NumberOfPixelsCounted++;

            if (this->m_NumberOfPixelsCounted == 1) {
                m_meanCurr_f = fixedValue;
                m_meanCurr_m = movingValue;
            } else {
                m_meanPrev_f = m_meanCurr_f;
                m_meanPrev_m = m_meanCurr_m;
                m_sPrev_f = m_sCurr_f;
                m_sPrev_m = m_sCurr_m;

                m_meanCurr_f = m_meanPrev_f + (fixedValue - m_meanPrev_f) / this->m_NumberOfPixelsCounted;
                m_sCurr_f = m_sPrev_f + (fixedValue - m_meanPrev_f) * (fixedValue - m_meanCurr_f);
                m_varianceCurr_f = m_sCurr_f / (this->m_NumberOfPixelsCounted - 1);

                m_meanCurr_m = m_meanPrev_m + (movingValue - m_meanPrev_m) / this->m_NumberOfPixelsCounted;
                m_sCurr_m = m_sPrev_m + (movingValue - m_meanPrev_m) * (movingValue - m_meanCurr_m);
                m_varianceCurr_m = m_sCurr_m / (this->m_NumberOfPixelsCounted - 1);
            }
        }

        ++ti;
    }

    if (this->m_SubtractMean && this->m_NumberOfPixelsCounted > 0) {
        sff -= (sf * sf / this->m_NumberOfPixelsCounted);
        smm -= (sm * sm / this->m_NumberOfPixelsCounted);
        sfm -= (sf * sm / this->m_NumberOfPixelsCounted);
    }

    RealType denom = -1.0 * sqrt(sff * smm);

    if (this->m_NumberOfPixelsCounted > 0 && denom != 0.0) {
        measure = sfm / denom;
    } else {
        measure = NumericTraits<MeasureType>::Zero;
    }

//    std::cout << "movingValueMin: " << movingValueMin << " movingValueMax: " << movingValueMax << std::endl;
//    std::cout << "fixedValueMin: " << fixedValueMin << " fixedValueMax: " << fixedValueMax << std::endl;
//    std::cout << "max coordinates: " << maxX << " " << maxX << " " << maxZ << std::endl;
//    std::cout << "min coordinates: " << minX << " " << minY << " " << minZ << std::endl;

    return measure;
}

template <typename TFixedImage, typename TMovingImage>
typename NormalizedCorrelationTwoImageToOneImageMetric<TFixedImage, TMovingImage>::MeasureType
NormalizedCorrelationTwoImageToOneImageMetric<TFixedImage, TMovingImage>::GetValue(
    const TransformParametersType& parameters) const
{
    if (!this->m_FixedImage1) {
        itkExceptionMacro(<< "Fixed image1 has not been assigned");
    }

    if (!this->m_FixedImage2) {
        itkExceptionMacro(<< "Fixed image2 has not been assigned");
    }

    if (!this->m_Filter1) {
        itkExceptionMacro(<< "Fixed filter1 has not been assigned");
    }

    if (!this->m_Filter2) {
        itkExceptionMacro(<< "Fixed filter1 has not been assigned");
    }

    if (!this->SetTransformParameters(parameters)) {
        std::cout << "Measure: 0.0, transform Invalid!" << std::endl;
        std::cout << "=======================================================" << std::endl;

        return 0.0;
    }
    return GetValue();
}

template <typename TFixedImage, typename TMovingImage>
typename NormalizedCorrelationTwoImageToOneImageMetric<TFixedImage, TMovingImage>::MeasureType
NormalizedCorrelationTwoImageToOneImageMetric<TFixedImage, TMovingImage>::GetValue() const
{

    if (!this->m_FixedImage1) {
        itkExceptionMacro(<< "Fixed image1 has not been assigned");
    }

    if (!this->m_FixedImage2) {
        itkExceptionMacro(<< "Fixed image2 has not been assigned");
    }

    if (!this->m_Filter1) {
        itkExceptionMacro(<< "Fixed filter1 has not been assigned");
    }

    if (!this->m_Filter2) {
        itkExceptionMacro(<< "Fixed filter1 has not been assigned");
    }

    // Calculate the measure value between fixed image 1 and the moving image
    auto oldprecision = std::cout.precision();


//    std::cout<<"Region "  <<this->GetFixedImageRegion1() <<std::endl;
//    std::cout<<"Buffered Region " <<this->GetFixedImage1()->GetBufferedRegion() <<std::endl;
//    std::cout<<"Buffered Region " <<this->m_Filter1->GetOutput()->GetBufferedRegion() <<std::endl;

    MeasureType measure1 = CalculateMeasure(1);
//    std::cout.precision(std::numeric_limits<double>::digits10 + 2);
//    std::cout << "Measure1: " << measure1 << " Counts: " << this->m_NumberOfPixelsCounted << std::endl;
//    std::cout.precision(oldprecision);

    MeasureType measure2 = CalculateMeasure(2);
//    std::cout.precision(std::numeric_limits<double>::digits10 + 2);
//    std::cout << "Measure2: " << measure2 << " Counts: " << this->m_NumberOfPixelsCounted << std::endl;

    auto measure = (measure1 + measure2) / 2.0;

//    std::cout << "Measure: " << measure << std::endl;
//    std::cout << "=======================================================" << std::endl;
//    std::cout.precision(oldprecision);

    return measure;
}

template <typename TFixedImage, typename TMovingImage>
void NormalizedCorrelationTwoImageToOneImageMetric<TFixedImage, TMovingImage>::GetDerivative(
    const TransformParametersType& itkNotUsed(parameters),
    DerivativeType& itkNotUsed(derivative)) const
{
    // under construction
}

template <typename TFixedImage, typename TMovingImage>
void NormalizedCorrelationTwoImageToOneImageMetric<TFixedImage, TMovingImage>::GetValueAndDerivative(
    const TransformParametersType& itkNotUsed(parameters),
    MeasureType& itkNotUsed(value),
    DerivativeType& itkNotUsed(derivative)) const
{
    // under construction
}

template <typename TFixedImage, typename TMovingImage>
void NormalizedCorrelationTwoImageToOneImageMetric<TFixedImage, TMovingImage>::PrintSelf(std::ostream& os,
    Indent indent) const
{
    Superclass::PrintSelf(os, indent);
    os << indent << "SubtractMean: " << m_SubtractMean << std::endl;
}

} // end namespace itk

#endif
