﻿#include "vtkContourTopogramProjectionFilter.h"


#include "vtkCallbackCommand.h"
#include "vtkCell.h"
#include "vtkCellArray.h"
#include "vtkCellData.h"
#include "vtkContourHelper.h"
#include "vtkContourValues.h"
#include "vtkGarbageCollector.h"
#include "vtkGenericCell.h"
#include "vtkInformation.h"
#include "vtkInformationVector.h"
#include "vtkNew.h"
#include "vtkObjectFactory.h"
#include "vtkPointData.h"
#include "vtkPolyData.h"
#include "vtkPolyDataNormals.h"
#include "vtkStreamingDemandDrivenPipeline.h"
#include "vtkTimerLog.h"
#include "vtkUniformGrid.h"
#include "vtkGraph.h"

#include "vtkTransformPolyDataFilter.h"
#include <cmath>

#include "vtkSmartPointer.h"
#define VTK_CREATE(type, name) vtkSmartPointer<type> name = vtkSmartPointer<type>::New()


vtkObjectFactoryNewMacro(vtkContourTopogramProjectionFilter);


vtkContourTopogramProjectionFilter::vtkContourTopogramProjectionFilter()
{

    this->dProjectionLPSOffset[0]=0.;
    this->dProjectionLPSOffset[1]=0.;
    this->dProjectionLPSOffset[2]=0.;

    this->dImportOffsetLPS[0] = 0.;
    this->dImportOffsetLPS[1] = 0.;
    this->dImportOffsetLPS[2] = 0.;

    this->dSCD = 0.;
    this->dPOffset = 0.;

    this->m_RefTransform = nullptr;
    this->m_Transform = nullptr;

    // by default process active point scalars
    this->SetInputArrayToProcess(
                0, 0, 0, vtkDataObject::FIELD_ASSOCIATION_POINTS, vtkDataSetAttributes::SCALARS);
}


vtkContourTopogramProjectionFilter::~vtkContourTopogramProjectionFilter()
{
    if (this->m_Transform)
     {
       this->m_Transform->UnRegister(this);
     }

    if (this->m_RefTransform)
     {
       this->m_RefTransform->UnRegister(this);
     }
}

void vtkContourTopogramProjectionFilter::cleanUpFilter(){

    this->dProjectionLPSOffset[0]=0.;
    this->dProjectionLPSOffset[1]=0.;
    this->dProjectionLPSOffset[2]=0.;

    this->dImportOffsetLPS[0] = 0.;
    this->dImportOffsetLPS[1] = 0.;
    this->dImportOffsetLPS[2] = 0.;

    this->dSCD = 0.;

    this->m_RefTransform = nullptr;
    this->m_Transform = nullptr;

}

int vtkContourTopogramProjectionFilter::RequestData(
        vtkInformation* request, vtkInformationVector** inputVector, vtkInformationVector* outputVector)
{

    // Get the info objects.
    vtkInformation* inInfo = inputVector[0]->GetInformationObject(0);
    vtkInformation* outInfo = outputVector->GetInformationObject(0);

    // Get the input and output.
    vtkPointSet* psInput = vtkPointSet::SafeDownCast(inInfo->Get(vtkDataObject::DATA_OBJECT()));
    vtkGraph* graphInput = vtkGraph::SafeDownCast(inInfo->Get(vtkDataObject::DATA_OBJECT()));
    vtkPolyData* output = vtkPolyData::SafeDownCast(outInfo->Get(vtkDataObject::DATA_OBJECT()));

    vtkPoints* points = nullptr;
    if (psInput)
    {
        points = psInput->GetPoints();
    }
    else
    {
        points = graphInput->GetPoints();
    }

    // check input
    // If no points, then nothing to do.
    if (points == nullptr)
    {
//        std::cout << "Cannot Project; no input points" << std::endl;
        vtkDebugMacro("Cannot Project; no input points");
        return 1;
    }

    // If reference transform, then nothing to do.
    if (m_RefTransform == nullptr)
    {
//        std::cout << "Cannot Project; no input reference projection transform" << std::endl;
        vtkDebugMacro("Cannot Project; no input reference projection transform");
        return 1;
    }

    // If transform, then nothing to do.
    if (m_Transform == nullptr)
    {
//        std::cout << "Cannot Project; no input projection transform" << std::endl;
        vtkDebugMacro("Cannot Project; no input projection transform");
        return 1;
    }

    vtkPoints* PrjPoints = vtkPoints::New();
    PrjPoints->SetNumberOfPoints( points->GetNumberOfPoints() );

    double pp0[3], pp1[3], pp2[3];

    for(vtkIdType ii = 0; ii < points->GetNumberOfPoints() ; ii++){

        /** input points are in LPS, should be corrected for import offset and mapped onto
         * the projection geometry origin */
        pp0[0] = points->GetPoint(ii)[0] + dProjectionLPSOffset [0] - dImportOffsetLPS[0];
        pp0[1] = points->GetPoint(ii)[1] + dProjectionLPSOffset [1] - dImportOffsetLPS[1];
        pp0[2] = points->GetPoint(ii)[2] + dProjectionLPSOffset [2] - dImportOffsetLPS[2];

        /** LPS points in projection can now be mapped to the
         * projection geometry using the inverse of the transform */
        m_Transform->GetInverse()->TransformPoint(pp0, pp1);

        /** correct x coordinate by the ratio of distance from source
         *  Basic intercept theorem */
        if(pp1[2] != 0.){
            pp1[0] = (pp1[0] * dSCD / abs(pp1[2])) /*+ dPOffset*/;
        } else {
            vtkDebugMacro("point at SCD distance from panel. Skipping lateral correction for this point.");
        }
        /** Go on the virtual detector */
        pp1[2] = -dSCD;

        /** Points can be mapped back to LPS coordinates */
        m_RefTransform->TransformPoint(pp1, pp2);

        /** and back to 'real' LPS */
        pp2[0] -= dProjectionLPSOffset [0];
        pp2[1] -= dProjectionLPSOffset [1];
        pp2[2] -= dProjectionLPSOffset [2];

        PrjPoints->InsertPoint(ii,pp2);
     }


    output->SetPoints( PrjPoints );

    vtkIdType numPoints = points->GetNumberOfPoints();

    VTK_CREATE(vtkCellArray, cells);
    if (cells->AllocateEstimate(numPoints, 1) == true){

    } else {
        std::cout << "vtkContourTopogramProjectionFilter Can't allocate cells " << std::endl;
        PrjPoints->Delete();
        return 1;
    }


    for (vtkIdType i = 0; i < numPoints; i++)
    {
        cells->InsertNextCell(1, &i);
    }

    output->SetVerts(cells);

    PrjPoints->Delete();

    output->Squeeze();

    return 1;


}

void vtkContourTopogramProjectionFilter::SetImportOffsetLPS(const double * dP)
{
    memcpy(dImportOffsetLPS,dP,3*sizeof(double));
}

void vtkContourTopogramProjectionFilter::SetSCD(const double dVal){
    dSCD = dVal;
}

void vtkContourTopogramProjectionFilter::SetPanelOffset(const double dVal){
    dPOffset = dVal;
}


void vtkContourTopogramProjectionFilter::SetLPStoProjectionLPSGeometryOffset(const double * dP)
{
    memcpy(dProjectionLPSOffset,dP,3*sizeof(double));
}

void vtkContourTopogramProjectionFilter::SetTransformMatrix( vtkMatrix4x4 *userMatrix)
{
    if(this->m_Transform == nullptr){
        m_Transform = vtkTransform::New();
    }
    m_Transform->SetMatrix(userMatrix);
    this->Modified();
}


void vtkContourTopogramProjectionFilter::SetReferenceTransformMatrix( vtkMatrix4x4 *userMatrix)
{
    if(this->m_RefTransform == nullptr){
        m_RefTransform = vtkTransform::New();
    }
    m_RefTransform ->SetMatrix(userMatrix);
    this->Modified();
}


void vtkContourTopogramProjectionFilter ::PrintSelf(ostream& os, vtkIndent indent)
{
    this->Superclass::PrintSelf(os, indent);
}


int vtkContourTopogramProjectionFilter::FillInputPortInformation(int, vtkInformation* info)
{
    info->Remove(vtkAlgorithm::INPUT_REQUIRED_DATA_TYPE());
    info->Append(vtkAlgorithm::INPUT_REQUIRED_DATA_TYPE(), "vtkGraph");
    info->Append(vtkAlgorithm::INPUT_REQUIRED_DATA_TYPE(), "vtkPointSet");
    return 1;
}
